#!/usr/bin/env python3

from os import path
import os
import json

import networkx as nx
import shapefile
from networkx import nx_agraph

import util_eddies

def read_eddy_graph(edgelist, shpc_dir=None, orientation="Cyclones"):
    if os.access(edgelist, os.R_OK):
        G = nx.read_edgelist(edgelist, create_using=nx.DiGraph, nodetype=int)
        dir_edgelist = path.dirname(edgelist)
        e_overestim_file = path.join(dir_edgelist, "e_overestim.txt")
        with open(e_overestim_file) as f:
            G.graph["e_overestim"] = int(f.read())

        if shpc_dir is not None:
            # Read and set attributes of eddies:
            SHPC = util_eddies.SHPC_class(shpc_dir, orientation)
            set_attribute(G, SHPC, orientation)
    else:
        print(edgelist, "not readable")
        G = nx.DiGraph()

    return G


def set_attribute(G, SHPC, orientation):
    reader = SHPC.get_reader(0, orientation, "extremum")

    if G.number_of_nodes() <= SHPC.get_n_shapes(0, orientation) / 100:
        for n in G:
            date_index, eddy_index = util_eddies.node_to_date_eddy(
                n, G.graph["e_overestim"]
            )
            ishape = SHPC.comp_ishape(date_index, eddy_index, 0, orientation)
            shape_rec = reader.shapeRecord(ishape)
            G.add_node(
                n,
                coordinates=tuple(shape_rec.shape.points[0]),
                speed=shape_rec.record.speed,
                ssh=shape_rec.record.ssh,
            )
    else:
        for shape_rec in reader:
            n = util_eddies.date_eddy_to_node(
                shape_rec.record.date,
                shape_rec.record.eddy_index,
                G.graph["e_overestim"],
            )

            if n in G:
                G.add_node(
                    n,
                    coordinates=tuple(shape_rec.shape.points[0]),
                    speed=shape_rec.record.speed,
                    ssh=shape_rec.record.ssh,
                )


def partition_graph_date(A, e_overestim):
    """Add subgraphs in pygraphviz graph, one subgraph for each date. We
    use the fact that the date can be deduced from the node
    identification number.

    """

    my_subgraphs = {}

    for n in A.iternodes():
        date_index = util_eddies.node_to_date_eddy(
            int(n), e_overestim, only_date=True
        )

        if date_index in my_subgraphs:
            my_subgraphs[date_index].append(n)
        else:
            my_subgraphs[date_index] = [n]

    for date, s in my_subgraphs.items():
        A.add_subgraph(s, rank="same")


def add_component(G: nx.DiGraph, source: int):
    """Add the weakly-connected component containing node source.

    The component is a set of nodes. Note this is not the same as the
    union of source, its descendants and ancestors.

    """

    if "component" in G.nodes[source]:
        print(f"add_component: component containing {source} already added")
    else:
        undirected_G = nx.to_undirected(G)
        node_generator = nx.dfs_preorder_nodes(undirected_G, source)
        component = set(node_generator)

        if "component_list" in G.graph:
            G.graph["component_list"].append(component)
        else:
            G.graph["component_list"] = [component]

        for n in component:
            G.nodes[n]["component"] = component


if __name__ == "__main__":
    import argparse

    # Parse arguments:
    parser = argparse.ArgumentParser()
    parser.add_argument("edgelist_file")
    parser.add_argument(
        "--gv", help="Create Graphviz file", action="store_true"
    )
    parser.add_argument(
        "-n", "--node", help="Select component containing node", type=int
    )
    parser.add_argument(
        "-m",
        "--merging",
        help="Report merging and splitting events",
        action="store_true",
    )
    args = parser.parse_args()

    G = read_eddy_graph(args.edgelist_file)

    if args.node is None:
        H = G
    else:
        add_component(G, args.node)
        H = G.subgraph(G.nodes[args.node]["component"])

    len_H = len(H)
    print("Number of nodes:", len_H)
    print(
        "(This should be the number of eddies minus the number of isolated "
        "nodes.)"
    )
    print("Number of edges:", H.number_of_edges())
    print(
        "Number of weakly connected components:",
        nx.number_weakly_connected_components(H),
    )
    n1 = len_H
    n2 = len_H

    if len_H >= 1:
        # First and last dates:
        node_index = min(G.nodes)
        my_date = util_eddies.node_to_date_eddy(
            node_index, G.graph["e_overestim"], only_date=True
        )
        print("First date:", my_date)
        node_index = max(G.nodes)
        my_date = util_eddies.node_to_date_eddy(
            node_index, G.graph["e_overestim"], only_date=True
        )
        print("Last date:", my_date)

    if args.merging:
        print("\nMerging:")

    for n, d in H.in_degree():
        if d >= 1:
            n2 -= 1
            if args.merging and d >= 2:
                print(list(H.predecessors(n)), "->", n)

    if args.merging:
        print("---")
        print("Splitting:")

    for n, d in H.out_degree():
        if d >= 1:
            n1 -= 1
            if args.merging and d >= 2:
                print(n, "->", list(H.successors(n)))

    if args.merging:
        print("---")

    print("Number of leaves:", n1)
    print("Number of roots:", n2)

    if args.gv:
        A = nx_agraph.to_agraph(H)
        partition_graph_date(A, H.graph["e_overestim"])
        A.write("eddy_graph.gv")
        print('Created file "eddy_graph.gv".')
